"""
Utility methods and classes for testing AvroAttrsBridge
"""
import io
import re

import attr
import fastavro

from openedx_events.bridge.avro_attrs_bridge_extensions import AvroAttrsBridgeExtention
from openedx_events.tooling import OpenEdxPublicSignal


def create_simple_signal(data_dict):
    """
    Create a basic OpenEdxPublicSignal with init_data = data_dict

    Arguments:
        data_dict: Description of attributes passed to the signal
    """
    return OpenEdxPublicSignal(
        event_type="simple.signal",
        data=data_dict
    )


def serialize_event_data_to_bytes(bridge, event_data):
    """
    Utility method to make sure an Avro serializer can actually serialize given a bridge schema and data
    to serialize

    Arguments:
        bridge: An instance of AvroAttrsBridge
        event_data: Event data to be sent via an OpenEdxPublicSignal's send_event method
    Returns:
        bytes: Byte representation of the event_data, to be sent over the wire
    """
    # Try to serialize using the generated schema.
    out = io.BytesIO()
    data_dict = bridge.to_dict(event_data)
    fastavro.schemaless_writer(out, bridge.schema_dict, data_dict)
    out.seek(0)
    return out.read()


def deserialize_bytes_to_event_data(bridge, bytes_from_wire):
    """
    Utility method to make sure an Avro deserializer can actually deserialize given a bridge and Avro-serialized
    data

    Arguments:
        bridge: an instance of AvroAttrsBridge
        bytes_from_wire: data that was serialized by an Avro serializer
    """
    data_file = io.BytesIO(bytes_from_wire)
    as_dict = fastavro.schemaless_reader(data_file, bridge.schema_dict)
    return bridge.from_dict(as_dict)


# Useful simple attr classes
@attr.s(auto_attribs=True)
class SimpleAttrs:
    """Class with all primitive type fields"""
    boolean_field: bool
    int_field: int
    float_field: float
    bytes_field: bytes
    string_field: str


@attr.s(auto_attribs=True)
class SubTestData0:
    """Subclass for testing nested attrs"""
    sub_name: str
    course_id: str


@attr.s(auto_attribs=True)
class SubTestData1:
    """Subclass for testing nested attrs"""
    sub_name: str
    course_id: str


@attr.s(auto_attribs=True)
class TestData:
    """More complex class for testing nested attrs"""
    sub_name: str
    course_id: str
    sub_test_0: SubTestData0
    sub_test_1: SubTestData1


@attr.s(frozen=True)
class SimpleAttrsWithDefaults:
    """Test attrs with nullable values"""
    boolean_field = attr.ib(type=bool, default=None)
    int_field = attr.ib(type=int, default=None)
    float_field = attr.ib(type=float, default=None)
    bytes_field = attr.ib(type=bytes, default=None)
    string_field = attr.ib(type=str, default=None)


@attr.s(frozen=True)
class NestedAttrsWithDefaults:
    """Test attrs with nullable values"""
    field_0 = attr.ib(type=SimpleAttrsWithDefaults, default=None)


class NonAttrs:
    """Data class not decorated with @attr. For testing bridge extension"""
    def __init__(self, val0, val1):
        self.val0 = val0
        self.val1 = val1

    def __eq__(self, other):
        # Treat all instances with the same values as equal for easier testing
        return self.val0 == other.val0 and self.val1 == other.val1


class SimpleBridgeExtension(AvroAttrsBridgeExtention):
    """
    Simple Bridge Extension for de/serializing
    """

    cls = NonAttrs

    def serialize(self, obj) -> str:
        """Serialize obj into string."""
        return f"{obj.val0}:{obj.val1}"

    def deserialize(self, data: str):
        """Deserialize string into obj."""
        bits = re.split(":", data)
        return NonAttrs(bits[0], bits[1])

    def record_fields(self):
        """Define Avro schema for self.cls."""
        return "string"



"""
Code to convert attr classes to Avro specification.

TODO (EventBus): handle optional parameters and allow for schema evolution (ARCHBOM-2013)
"""
import json
from typing import Any, Dict

import attr
import fastavro

from openedx_events.bridge.avro_attrs_bridge_extensions import (
    CourseKeyAvroAttrsBridgeExtension,
    DatetimeAvroAttrsBridgeExtension,
)
from openedx_events.bridge.avro_types import PYTHON_TYPE_TO_AVRO_MAPPING


class AvroAttrsBridge:
    """
    Convert between Avro and OpenEdxPublicSignal data specifications.

    Intended use: To abstract serialization and deserialization of openedx-events to send over an event bus
    The bridge can be used to both automatically generate an Avro schema for the events sent by the associated
    signal instance and to serialize the events themselves using the generated schema.
    TODO (EventBus): rename this class and file to convey that the bridge now works with entire signal
    definitions rather than just attrs classes (ARCHBOM-2101)
    """

    # default extensions, can be overwritten by passing in extensions during obj initialization
    default_extensions = {
        DatetimeAvroAttrsBridgeExtension.cls: DatetimeAvroAttrsBridgeExtension(),
        CourseKeyAvroAttrsBridgeExtension.cls: CourseKeyAvroAttrsBridgeExtension(),
    }

    def __init__(self, signal, extensions=None):
        """
        Init method for Avro Attrs Bridge.

        Arguments:
            signal: An instance of OpenEdxPublicSignal
            extensions: dict mapping a Class to an instance of the required AvroAttrsBridgeExtension subclass, eg
                       { MyDataClass : MyDataClassAvroAttrsBridgeExtension() }
        """
        self.extensions = {}
        self.extensions.update(self.default_extensions)
        if isinstance(extensions, dict):
            self.extensions.update(extensions)

        self._signal = signal
        # Used by _create_avro_field_definition function to keep track of which
        # records have already been defined in schema.
        # Reason: fastavro does not allow you to define record with same name twice
        self.schema_record_names = set()
        self.schema_dict = self._avro_schema_dict_from_signal()

        # make sure the schema is parsable
        fastavro.parse_schema(self.schema_dict)

    def schema_str(self):
        """Json dumps schema dict into a string."""
        return json.dumps(self.schema_dict, sort_keys=True)

    def _avro_schema_dict_from_signal(self):
        """
        Generate the Avro schema for events sent by self._signal.

        TODO (EventBus): Include required CloudEvent fields with sensible defaults
        Returns:
            complex dict that defines Avro schema for events sent by self._signal
        """
        base_schema = {
            "name": "CloudEvent",
            "type": "record",
            "doc": "Avro Event Format for CloudEvents created with openedx_events/avro_attrs_bridge",
            "fields": [],
        }

        for data_key, data_type in self._signal.init_data.items():
            base_schema["fields"].append(self._create_avro_field_definition(data_key, data_type))
        return base_schema

    def _create_avro_field_definition(self, data_key, data_type, default_is_none=False):
        """
        Create an Avro schema field definition from an OpenEdxPublicSignal data definition.

        Arguments:
            data_key: Field name
            data_type: Python data type, eg `str`, `CourseKey`, `CourseEnrollmentData`
            default_is_none: boolean indicating whether this field has 'None' as a default
        """
        field = {"name": data_key}

        # Case 1: data_type has known extension
        if extension := self.extensions.get(data_type, None):
            field["type"] = extension.record_fields()
        # Case 2: data_type is a simple type that can be converted directly to an Avro type
        elif data_type in PYTHON_TYPE_TO_AVRO_MAPPING:
            if PYTHON_TYPE_TO_AVRO_MAPPING[data_type] in ["record", "array"]:
                # TODO (EventBus): figure out how to handle container types (dicts and arrays). (ARCHBOM-2095)
                raise Exception("Unable to generate Avro schema for dict or array fields")
            avro_type = PYTHON_TYPE_TO_AVRO_MAPPING[data_type]
            field["type"] = avro_type

        # Case 2: data_type is an attrs class
        elif hasattr(data_type, "__attrs_attrs__"):
            # Inner Attrs Class

            # fastavro does not allow you to redefine the same record type more than once,
            # so only define an attr record once
            if data_type.__name__ in self.schema_record_names:
                field["type"] = data_type.__name__
            else:
                self.schema_record_names.add(data_type.__name__)
                field = self._generate_avro_record_for_attrs_class(
                    data_type, data_key
                )
        else:
            raise TypeError(
                f"Data type {data_type} is not supported by AvroAttrsBridge. The data type needs to either"
                " be one of the types in PYTHON_TYPE_TO_AVRO_MAPPING, an attrs decorated class, or one of the types"
                " defined in self.extensions."
            )
        if default_is_none:
            field["default"] = "null"
            single_type = field["type"]
            field["type"] = ["null", single_type]
        return field

    def _generate_avro_record_for_attrs_class(
        self, attrs_class, field_name: str
    ) -> Dict[str, Any]:
        """
        Generate Avro record for attrs_class.

        Will also recursively generate Avro records for any sub attr classes and any custom types.

        Custom types are handled by AvroAttrsBridgeExtension subclass instance values defined in self.extensions dict
        """
        field: Dict[str, Any] = {}
        field["name"] = field_name
        record_type = dict(name=attrs_class.__name__, type="record", fields=[])

        for attribute in attrs_class.__attrs_attrs__:
            record_type["fields"].append(
                self._create_avro_field_definition(attribute.name, attribute.type, attribute.default is None)
            )
        field["type"] = record_type
        return field

    def to_dict(self, event_data):
        """
        Convert event_data into dictionary that matches Avro schema (self.schema).

        Warning: this does not validate that the data_dict input matches self._signal

        Arguments:
            event_data: dict with all the values specified in OpenEdxPublicSignal.init_data.
        """
        # TODO (EventBus): is there better way to do this besides using json?
        # This first converts data_dict to json string and then back to dict.

        return json.loads(
            json.dumps(
                event_data, sort_keys=True, default=self._event_value_to_json
            )
        )

    def _event_value_to_json(self, value):
        """
        Serialize a top-level value in an event data dictionary to match the Avro schema.
        """
        # Case 1: Value is an instance of an attrs-decorated class
        if hasattr(value, "__attrs_attrs__"):
            return attr.asdict(value, value_serializer=self._serialize_non_attrs_instance)
        # Case 2: Value is an instance of a class (or subclass) for which the bridge has a known extension
        for extended_class, extension in self.extensions.items():
            if issubclass(type(value), extended_class):
                return extension.serialize(value)
        # Case 3: Default
        return value

    def _serialize_non_attrs_instance(self, _, field, value):
        """
        Use an extension to serialize a value of the appropriate class.

        Used as a callback to attr.asdict to handle any inner fields that are not attrs classes or primitives
        """
        extension = self.extensions.get(field.type, None)
        if extension is not None:
            return extension.serialize(value)
        return value

    def from_dict(self, data: dict):
        """
        Convert dict into event data that can be sent with self._signal.

        Arguments:
            data: Dictionary returned from AvroDeserializer

        Returns:
            dict: Event data dictionary
        """
        return dict([(data_key, self._deserialized_avro_dict_to_object(data[data_key], data_type))
                     for data_key, data_type in self._signal.init_data.items()])

    def _deserialized_avro_dict_to_object(self, data: dict, data_type):
        """
        Convert dictionary entry into an instance of data_type.

        Used to convert messages from an AvroDeserializer into events that can be sent by the
        appropriate signal instance

        Arguments:
            data: Dictionary returned from AvroDeserializer
            data_type: Desired Python data type, eg `str`, `CourseKey`, `CourseEnrollmentData`

        Returns:
            An instance of data_type
        """
        if not hasattr(data_type, '__attrs_attrs__'):
            if extension := self.extensions.get(data_type, None):
                return extension.deserialize(data)
            else:
                raise TypeError(
                    f"Unable to deserialize {data_type} data, please add extension for custom data type"
                )

        for attribute in data_type.__attrs_attrs__:
            if attribute.name in data:
                sub_data = data[attribute.name]
                if sub_data is None:
                    # delete keys that have defaults in attr class and which have None as value
                    # this is to let attr class take care of creating default values
                    if attribute.default is not attr.NOTHING:
                        del data[attribute.name]
                else:
                    if hasattr(attribute.type, "__attrs_attrs__"):
                        if attribute.name in data:
                            data[attribute.name] = self._deserialized_avro_dict_to_object(
                                sub_data, attribute.type
                            )
                    elif attribute.type in self.extensions:
                        extension = self.extensions.get(attribute.type)
                        if attribute.name in data:
                            data[attribute.name] = extension.deserialize(sub_data)
                        else:
                            raise Exception(
                                f"Necessary key: {attribute.name} not found in data dict"
                            )
                    elif attribute.type not in PYTHON_TYPE_TO_AVRO_MAPPING:
                        raise TypeError(
                            f"Unable to deserialize {attribute.type} data, please add extension for custom data type"
                        )

        return data_type(**data)



"""
Tests for AvroAttrsBridge.
"""
from datetime import datetime
from unittest import TestCase

from opaque_keys.edx.keys import CourseKey

from openedx_events.bridge.avro_attrs_bridge import AvroAttrsBridge
from openedx_events.bridge.tests.test_utilities import (
    NestedAttrsWithDefaults,
    NonAttrs,
    SimpleAttrs,
    SimpleAttrsWithDefaults,
    SimpleBridgeExtension,
    SubTestData0,
    SubTestData1,
    TestData,
    create_simple_signal,
    deserialize_bytes_to_event_data,
    serialize_event_data_to_bytes,
)


class TestAvroAttrsBridge(TestCase):
    """
    Test AvroAttrsBridge functionality
    """

    def test_simple_schema_generation(self):
        SIGNAL = create_simple_signal({"event_data": SimpleAttrs})

        bridge = AvroAttrsBridge(SIGNAL)
        expected_dict = {
            "type": "record",
            "name": "CloudEvent",
            "doc": "Avro Event Format for CloudEvents created with openedx_events/avro_attrs_bridge",
            "fields": [
                {"name": "event_data", "type":
                    {"name": "SimpleAttrs", "type": "record", "fields": [
                        {"name": "boolean_field", "type": "boolean"},
                        {"name": "int_field", "type": "long"},
                        {"name": "float_field", "type": "double"},
                        {"name": "bytes_field", "type": "bytes"},
                        {"name": "string_field", "type": "string"},
                    ]
                     }
                 }
            ],
        }

        self.assertDictEqual(bridge.schema_dict, expected_dict)

    def test_nested_attrs_object_serialization(self):
        SIGNAL = create_simple_signal({"test_data": TestData})
        bridge = AvroAttrsBridge(SIGNAL)

        expected_dict = {
            "type": "record",
            "name": "CloudEvent",
            "doc": "Avro Event Format for CloudEvents created with openedx_events/avro_attrs_bridge",
            "fields": [
                {"name": "test_data", "type":
                    {"name": "TestData", "type": "record", "fields": [
                        {"name": "sub_name", "type": "string"},
                        {"name": "course_id", "type": "string"},
                        {"name": "sub_test_0", "type": {
                            "name": "SubTestData0",
                            "type": "record",
                            "fields": [
                                {"name": "sub_name", "type": "string"},
                                {"name": "course_id", "type": "string"},
                            ]
                        }},
                        {"name": "sub_test_1", "type": {
                            "name": "SubTestData1",
                            "type": "record",
                            "fields": [
                                {"name": "sub_name", "type": "string"},
                                {"name": "course_id", "type": "string"},
                            ]
                        }},
                    ]}
                 },
            ],
        }

        self.assertDictEqual(bridge.schema_dict, expected_dict)

    def test_multiple_top_level_fields(self):
        SIGNAL = create_simple_signal({
            "top_level_key_0": SubTestData0,
            "top_level_key_1": SubTestData1,
        })
        bridge = AvroAttrsBridge(SIGNAL)

        expected_dict = {
            "type": "record",
            "name": "CloudEvent",
            "doc": "Avro Event Format for CloudEvents created with openedx_events/avro_attrs_bridge",
            "fields": [
                {"name": "top_level_key_0", "type":
                    {
                        "name": "SubTestData0",
                        "type": "record",
                        "fields": [
                            {"name": "sub_name", "type": "string"},
                            {"name": "course_id", "type": "string"},
                        ]
                    }
                 },
                {"name": "top_level_key_1", "type":
                    {
                        "name": "SubTestData1",
                        "type": "record",
                        "fields": [
                            {"name": "sub_name", "type": "string"},
                            {"name": "course_id", "type": "string"},
                        ]
                    }
                 },
            ],
        }
        self.assertDictEqual(bridge.schema_dict, expected_dict)

    def test_convert_event_data_to_dict(self):
        """
        Tests that an event with complex attrs objects can be converted to dict and back
        """

        SIGNAL = create_simple_signal({"test_data": TestData})
        bridge = AvroAttrsBridge(SIGNAL)
        # A test record that we can try to serialize to avro.
        test_data = TestData(
            "foo",
            "bar.course",
            SubTestData0("a.sub.name", "a.nother.course"),
            SubTestData1("b.uber.sub.name", "b.uber.another.course"),
        )

        data_dict = bridge.to_dict({"test_data": test_data})
        expected_dict = {
            "test_data": {
                "course_id": "bar.course",
                "sub_name": "foo",
                "sub_test_0": {"course_id": "a.nother.course", "sub_name": "a.sub.name"},
                "sub_test_1": {
                    "course_id": "b.uber.another.course",
                    "sub_name": "b.uber.sub.name",
                },
            }
        }
        self.assertDictEqual(data_dict, expected_dict)

    def test_convert_dict_to_event_data(self):
        initial_dict = {
            "test_data": {
                "course_id": "bar.course",
                "sub_name": "foo",
                "sub_test_0": {"course_id": "a.nother.course", "sub_name": "a.sub.name"},
                "sub_test_1": {
                    "course_id": "b.uber.another.course",
                    "sub_name": "b.uber.sub.name",
                },
            }
        }

        SIGNAL = create_simple_signal({"test_data": TestData})
        bridge = AvroAttrsBridge(SIGNAL)
        event_data = bridge.from_dict(initial_dict)

        test_data = event_data["test_data"]
        self.assertIsInstance(test_data, TestData)
        self.assertEqual(test_data.course_id, "bar.course")
        self.assertEqual(test_data.sub_name, "foo")

        sub_0 = test_data.sub_test_0
        self.assertIsInstance(sub_0, SubTestData0)
        self.assertEqual(sub_0.course_id, "a.nother.course")
        self.assertEqual(sub_0.sub_name, "a.sub.name")

        sub_1 = test_data.sub_test_1
        self.assertIsInstance(sub_1, SubTestData1)
        self.assertEqual(sub_1.course_id, "b.uber.another.course")
        self.assertEqual(sub_1.sub_name, "b.uber.sub.name")

    def test_full_serialize_and_deserialize(self):
        SIGNAL = create_simple_signal({"test_data": TestData})
        bridge = AvroAttrsBridge(SIGNAL)
        test_data = {"test_data": TestData(
            "foo",
            "bar.course",
            SubTestData0("a.sub.name", "a.nother.course"),
            SubTestData1("b.uber.sub.name", "b.uber.another.course"),
        )}

        as_bytes = serialize_event_data_to_bytes(bridge, test_data)
        deserialized = deserialize_bytes_to_event_data(bridge, as_bytes)
        self.assertEqual(test_data, deserialized)
        # check we can emit the deserialized event with the original signal
        SIGNAL.send_event(**deserialized)

    def test_default_datetime_extension_serialization(self):
        SIGNAL = create_simple_signal({"birthday": datetime})
        bridge = AvroAttrsBridge(SIGNAL)
        birthday = datetime(year=1989, month=9, day=6)
        test_data = {"birthday": birthday}
        data_dict = bridge.to_dict(test_data)
        assert data_dict == {"birthday": birthday.isoformat()}

    def test_default_datetime_extension_deserialization(self):
        SIGNAL = create_simple_signal({"birthday": datetime})
        bridge = AvroAttrsBridge(SIGNAL)
        birthday = datetime(year=1989, month=9, day=6)
        as_dict = {"birthday": birthday.isoformat()}
        event_data = bridge.from_dict(as_dict)
        birthday_deserialized = event_data["birthday"]
        self.assertIsInstance(birthday_deserialized, datetime)
        self.assertEqual(birthday_deserialized.year, 1989)
        self.assertEqual(birthday_deserialized.month, 9)
        self.assertEqual(birthday_deserialized.day, 6)

    def test_default_coursekey_extension_serialization(self):
        SIGNAL = create_simple_signal({"course": CourseKey})
        bridge = AvroAttrsBridge(SIGNAL)
        course_key = CourseKey.from_string("course-v1:edX+DemoX.1+2014")
        test_data = {"course": course_key}
        data_dict = bridge.to_dict(test_data)
        assert data_dict == {"course": str(course_key)}

    def test_default_coursekey_extension_deserialization(self):
        SIGNAL = create_simple_signal({"course": CourseKey})
        bridge = AvroAttrsBridge(SIGNAL)
        course_key = CourseKey.from_string("course-v1:edX+DemoX.1+2014")
        as_dict = {"course": str(course_key)}
        event_data = bridge.from_dict(as_dict)
        course_deserialized = event_data["course"]
        self.assertIsInstance(course_deserialized, CourseKey)
        self.assertEqual(course_deserialized, course_key)

    def test_custom_serializer(self):
        SIGNAL = create_simple_signal({"test_data": NonAttrs})
        bridge = AvroAttrsBridge(SIGNAL, extensions={SimpleBridgeExtension.cls: SimpleBridgeExtension()})
        test_data = {
            "test_data": NonAttrs("a.val", "a.nother.val")
        }
        as_bytes = serialize_event_data_to_bytes(bridge, test_data)
        self.assertEqual(test_data, deserialize_bytes_to_event_data(bridge, as_bytes))

    def test_throw_exception_on_unextended_custom_type(self):
        class UnextendedClass:
            pass

        SIGNAL = create_simple_signal({"unextended_class": UnextendedClass})
        with self.assertRaises(TypeError):
            AvroAttrsBridge(SIGNAL)

    def test_dict_to_event_data_fails_if_missing_fields(self):
        # missing "sub_name" field
        initial_dict = {
            "test_data": {
                "course_id": "bar.course",
            }
        }

        SIGNAL = create_simple_signal({"test_data": SubTestData0})
        bridge = AvroAttrsBridge(SIGNAL)
        with self.assertRaises(TypeError):
            _ = bridge.from_dict(initial_dict)

    def test_throw_exception_to_list_or_dict_types(self):
        LIST_SIGNAL = create_simple_signal({"list_input": list})
        DICT_SIGNAL = create_simple_signal({"list_input": dict})
        with self.assertRaises(Exception):
            AvroAttrsBridge(LIST_SIGNAL)

        with self.assertRaises(Exception):
            AvroAttrsBridge(DICT_SIGNAL)

    def test_optional_fields(self):
        SIGNAL = create_simple_signal({
            "data": SimpleAttrsWithDefaults
        })
        bridge = AvroAttrsBridge(SIGNAL)
        event_data = {"data": SimpleAttrsWithDefaults()}
        as_bytes = serialize_event_data_to_bytes(bridge, event_data)
        deserialized = deserialize_bytes_to_event_data(bridge, as_bytes)
        simple_attrs = deserialized["data"]
        self.assertIsNone(simple_attrs.boolean_field)
        self.assertIsNone(simple_attrs.int_field)
        self.assertIsNone(simple_attrs.float_field)
        self.assertIsNone(simple_attrs.bytes_field)
        self.assertIsNone(simple_attrs.string_field)

    def test_nested_optional_fields(self):
        SIGNAL = create_simple_signal({
            "data": NestedAttrsWithDefaults
        })
        bridge = AvroAttrsBridge(SIGNAL)
        event_data = {"data": NestedAttrsWithDefaults()}
        as_bytes = serialize_event_data_to_bytes(bridge, event_data)
        deserialized = deserialize_bytes_to_event_data(bridge, as_bytes)
        simple_attrs = deserialized["data"]
        self.assertIsNone(simple_attrs.field_0)
